#version 330 core

uniform mat4 world_model;
uniform mat4 view_world;
uniform mat4 clip_view;

in vec3 model_position;
in vec3 model_normal;
in vec3 model_tangent;
in vec3 color;
in vec2 texture;

out vec3 material_color;
out vec3 world_position;
out vec3 world_normal;
out vec3 world_tangent;
out vec3 world_bitangent;
out vec2 tex_coords;

void main()
{
   material_color = color;
   tex_coords = texture;
   
   gl_Position = world_model * vec4(model_position, 1);
   world_position = gl_Position.xyz;
   world_normal = (world_model * vec4(model_normal, 0)).xyz; //to be extended for asymmetric scaling
   world_tangent = (world_model * vec4(model_tangent, 0)).xyz; //to be extended for asymmetric scaling
   world_tangent -= dot(world_tangent, world_normal) * world_normal;
   world_bitangent = cross(world_normal, world_tangent);
   gl_Position = clip_view * view_world * gl_Position;
}

###

#version 330 core

#define PI 3.1415926538
#define MAX_NUM_LIGHT 10

struct AmbientLight
{
   vec3 color;
};
struct DirectionalLight
{
   vec3 color;
   vec3 direction;
};
struct PointLight
{
   vec3 color;
   vec3 position;
   vec3 direction;
   float inner;
   float outer;
};
struct Material
{
   float reflectivity;
   float shininess;
   sampler2D texture;
   sampler2D surface;
};

uniform int num_amb_lights;
uniform AmbientLight amb_lights[MAX_NUM_LIGHT];
uniform int num_dir_lights;
uniform DirectionalLight dir_lights[MAX_NUM_LIGHT];
uniform int num_poi_lights;
uniform PointLight poi_lights[MAX_NUM_LIGHT];
uniform vec3 world_cam_position;

uniform Material material;
in vec3 material_color;

in vec3 world_position;
in vec3 world_normal;
in vec3 world_tangent;
in vec3 world_bitangent;
in vec2 tex_coords;

out vec4 frag_color;

void main()
{
   vec3 world_normal_unit = normalize(world_normal);
   if(world_tangent != vec3(0) || world_bitangent != vec3(0))
   {
      vec3 tex_normal = normalize(texture2D(material.surface, tex_coords).rgb - vec3(0.5));
      world_normal_unit = tex_normal.r * normalize(world_tangent) + tex_normal.g * normalize(world_bitangent) + tex_normal.b * world_normal_unit;
   }
   
   vec3 camera_unit = normalize(world_cam_position - world_position);
   vec3 ambient_color = vec3(0.0);
   vec3 diffuse_color = vec3(0.0);
   vec3 specular_color = vec3(0.0);

   if(num_amb_lights + num_dir_lights + num_poi_lights == 0) ambient_color = vec3(1.0);
   for(int i = 0; i < num_amb_lights; i++) ambient_color += amb_lights[i].color;
   for(int i = 0; i < num_dir_lights; i++)
   {
      vec3 direction_unit = normalize(dir_lights[i].direction);
      float fraction = dot(world_normal_unit, -direction_unit);
      if(fraction <= 0) continue; //light from behind
      diffuse_color += fraction * dir_lights[i].color;
      specular_color += material.reflectivity * pow(max(0.0, dot(reflect(direction_unit, world_normal_unit), camera_unit)), material.shininess) * dir_lights[i].color;
   }
   for(int i = 0; i < num_poi_lights; i++)
   {
      vec3 direction = world_position - poi_lights[i].position;
      vec3 direction_unit = normalize(direction);
      float fraction = dot(world_normal_unit, -direction_unit);
      if(fraction <= 0) continue; //light from behind
      //intensity decreases with distance
      float distance = length(direction);
      float intensity = 1 / (distance * distance + 0.1);
      //intensity vanishes outside the light cone
      float cos_theta = dot(normalize(poi_lights[i].direction), direction_unit);
      intensity *=
         cos_theta >= poi_lights[i].inner ?
            1.0 :
         cos_theta >= poi_lights[i].outer ?
            -cos((poi_lights[i].outer - cos_theta) / (poi_lights[i].outer - poi_lights[i].inner) * PI) / 2 + 0.5
            : 0.0;
      
      //diffuse intensity decreases with inclination angle
      diffuse_color += intensity * fraction * poi_lights[i].color;
      //specular light
      specular_color += intensity * material.reflectivity * pow(max(0.0, dot(reflect(direction_unit, world_normal_unit), camera_unit)), material.shininess) * poi_lights[i].color;
   }
   //the material has either a color or a texture
   frag_color = (vec4(material_color, 0) + texture2D(material.texture, tex_coords)) * vec4(ambient_color + diffuse_color + specular_color, 1);
}